﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Text.RegularExpressions;
using Episerver.Marketing.Common.Helpers;
using EPiServer.ConnectForCampaign.Core.Configuration;
using EPiServer.ConnectForCampaign.Core.Helpers;
using EPiServer.ConnectForCampaign.Core.Implementation;
using EPiServer.ConnectForCampaign.Services;
using EPiServer.ConnectForCampaign.Services.Recipient;
using EPiServer.ConnectForCampaign.Services.Implementation;
using EPiServer.Framework.Cache;
using EPiServer.Framework.Localization;
using EPiServer.Logging;
using EPiServer.ServiceLocation;

namespace AlloySample.Business.Initialization
{
    public class CustomRecipientService : RecipientService
    {
        private const string EmailFormat = @"[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?";

        private static readonly ILogger _logger = LogManager.GetLogger(typeof(CustomRecipientService));
        private readonly LocalizationService _localizationService;
        private readonly IServiceClientFactory _serviceClientFactory;
        private readonly ICampaignConfig _campaignConfig;
        private readonly ICampaignSettings _campaignSettings;
        private readonly ICacheService _cacheService;
        private readonly ICookieHelper _cookieHelper;

        //private readonly string _recipientDetailsCacheKey = "EPiServer.ConnectForCampaign:RecipientDetails";

        private enum RecipientAddResponse
        {
            RecipientAdded = 0,
            ValidationFailed = 1,
            RecipientUnsubscribed = 2,
            RecipientBlacklisted = 3,
            RecipientBounceOverflowed = 4,
            RecipientAlreadyExists = 5,
            RecipientHasBeenFiltered = 6,
            GeneralError = 7
        }

        public CustomRecipientService(LocalizationService localizationService, IServiceClientFactory serviceClientFactory, IAuthenticationService authenticationService,
            ICampaignConfig campaignConfig, ICacheService cacheService, ICookieHelper cookieHelper, ICampaignSettings campaignSettings)
            : base(localizationService, serviceClientFactory, authenticationService, campaignConfig, cacheService, cookieHelper, campaignSettings)
        {
            _localizationService = localizationService;
            _serviceClientFactory = serviceClientFactory;
            _campaignConfig = campaignConfig;
            _cacheService = cacheService;
            _cookieHelper = cookieHelper;
            _campaignSettings = campaignSettings;
        }

        public override string UpsertRecipient(Dictionary<string, string> profileValues, long recipientListId, long optinProcessId, out string validationMessage)
        {
            validationMessage = string.Empty;
            var userId = string.Empty;

            //Original code. Failed to validate the present of email address if email field is named differently.
            //var email = profileValues.FirstOrDefault(kv => kv.Key.Equals("email", StringComparison.InvariantCultureIgnoreCase));

            //Workaround.
            var email = profileValues.FirstOrDefault(kv => Regex.IsMatch(kv.Value, EmailFormat, RegexOptions.IgnoreCase));

            var trackingCookie = _cookieHelper.GetTrackingCookie(Constants.ConnectorId.ToString(), Constants.DefaultConnectorInstanceId.ToString());
            var extractedEmailFromCookie = trackingCookie?.Where(cd => cd.DatasourceId == recipientListId.ToString()).FirstOrDefault()?.EntityId;
            var emailIsEmpty = email.Equals(default(KeyValuePair<string, string>));

            // Check whether we should update recipient or add new one.
            // We should update recipient when all the following is true:
            //   - config forceUpdateRecipient = true
            //   - Email or extractedEmailFromCookie is not empty
            //   - Recipient already exists in Campaign System (check using email)
            // or
            //   - config forceUpdateRecipient = false
            //   - extractedEmailFromCookie is not null
            //   - One of the following is true:
            //     + Email is empty
            //     + Email is not empty and email is the same as extractedEmailFromCookie
            //   - Recipient still exists in Campaign System (check using tracking cookie)
            var shouldUpdateRecipient = ((_campaignConfig.ForceUpdateRecipient && ((!emailIsEmpty && IsRecipientExisting(recipientListId, email.Value)) || (emailIsEmpty && !string.IsNullOrEmpty(extractedEmailFromCookie) && IsRecipientExisting(recipientListId, extractedEmailFromCookie)))) ||
                                    (!_campaignConfig.ForceUpdateRecipient && !string.IsNullOrEmpty(extractedEmailFromCookie)
                                    && (emailIsEmpty || email.Value.Equals(extractedEmailFromCookie, StringComparison.InvariantCultureIgnoreCase))
                                    && IsRecipientExisting(recipientListId, extractedEmailFromCookie)));

            if (shouldUpdateRecipient)
            {
                userId = !emailIsEmpty ? email.Value : extractedEmailFromCookie;

                var updateSuccess = UpdateRecipient(profileValues, recipientListId, userId, userId);
                if (!updateSuccess)
                {
                    _logger.Error($"Error updating recipient. userId: {userId}, recipientListId: {recipientListId}");
                    validationMessage = _localizationService.GetString("/episerver/campaign/servicevalidation/generalerror");
                    return string.Empty;
                }
                // clear cache for the recipient
                _cacheService.Remove(BuildCacheKey(recipientListId, userId));

                return userId;
            }

            if (emailIsEmpty)
            {
                validationMessage = _localizationService.GetString("/episerver/campaign/servicevalidation/emailrequired");
                _logger.Error(validationMessage);
                return string.Empty;
            }

            userId = email.Value;

            var apiResponse = InsertRecipient(profileValues, recipientListId, optinProcessId, userId, userId);
            var generalError = _localizationService.GetString("/episerver/campaign/servicevalidation/generalerror");

            if (apiResponse != null)
            {
                switch (apiResponse)
                {
                    case RecipientAddResponse.RecipientAdded:
                        // clear cache for the recipient
                        _cacheService.Remove(BuildCacheKey(recipientListId, userId));
                        _logger.Debug($"Recipient has been added: {userId}");
                        break;
                    case RecipientAddResponse.ValidationFailed:
                        _logger.Error($"Recipient validation failed. userId: {userId} - apiResponse: {(int)apiResponse}");
                        validationMessage = _localizationService.GetString("/episerver/campaign/servicevalidation/validationfailed");
                        return string.Empty;
                    case RecipientAddResponse.RecipientUnsubscribed:
                        _logger.Error($"Recipient is unsubscribed. userId: {userId} - apiResponse: {(int)apiResponse}");
                        validationMessage = generalError;
                        return string.Empty;
                    case RecipientAddResponse.RecipientBlacklisted:
                        _logger.Error($"Recipient is blacklisted. userId: {userId} - apiResponse: {(int)apiResponse}");
                        validationMessage = generalError;
                        return string.Empty;
                    case RecipientAddResponse.RecipientBounceOverflowed:
                        _logger.Error($"Recipient is bounce overflowed. userId: {userId} - apiResponse: {(int)apiResponse}");
                        validationMessage = generalError;
                        return string.Empty;
                    case RecipientAddResponse.RecipientAlreadyExists:
                        _logger.Error($"Recipient already exists. userId: {userId} - apiResponse: {(int)apiResponse}");
                        validationMessage = generalError;
                        return string.Empty;
                    case RecipientAddResponse.RecipientHasBeenFiltered:
                        _logger.Error($"Recipient has been filtered. userId: {userId} - apiResponse: {(int)apiResponse}");
                        validationMessage = generalError;
                        return string.Empty;
                    case RecipientAddResponse.GeneralError:
                        _logger.Error($"Error from InsertRecipient. userId: {userId} - apiResponse: {(int)apiResponse}");
                        validationMessage = generalError;
                        return string.Empty;
                    default:
                        _logger.Error($"Result was: {(int)apiResponse}");
                        break;
                }
            }

            return userId;
        }

        private bool IsRecipientExisting(long recipientListId, string userId)
        {
            try
            {
                var token = GetSessionToken();
                var recipientClient = _serviceClientFactory.GetRecipientClient();
                return recipientClient.contains(token, recipientListId, userId);
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message, ex);
                return false;
            }
        }

        /// <summary>
        /// Update Recipient to Campaign system
        /// </summary>
        /// <param name="profileValues"></param>
        /// <param name="recipientListId">Id of recipientList to update</param>
        /// <param name="userId"></param>
        /// <param name="email"></param>
        /// <returns>Return true if update success, otherwise return false</returns>
        private bool UpdateRecipient(Dictionary<string, string> profileValues, long recipientListId, string userId, string email)
        {
            var success = true;

            try
            {
                var token = GetSessionToken();
                var recipientClient = _serviceClientFactory.GetRecipientClient();

                var parsedValues = ParseProfileValues(profileValues);
                var request = new setAttributesRequest(token, recipientListId, userId, parsedValues.Item1, parsedValues.Item2);
                recipientClient.setAttributes(request);
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message, ex);
                success = false;
            }

            return success;
        }

        /// <summary>
        /// Add recipient to a recipient list
        /// </summary>
        /// <param name="profileValues"></param>
        /// <param name="recipientListId"></param>
        /// <param name="userId"></param>
        /// <param name="email"></param>
        /// <returns></returns>
        private RecipientAddResponse? InsertRecipient(Dictionary<string, string> profileValues, long recipientListId, long optinProcessId, string userId, string email)
        {
            try
            {
                var token = GetSessionToken();
                var recipientClient = _serviceClientFactory.GetRecipientClient();

                var parsedValues = ParseProfileValues(profileValues);
                var request = new add2Request(token, recipientListId, optinProcessId, userId, email, parsedValues.Item1, parsedValues.Item2);
                var response = recipientClient.add2(request);
                return (RecipientAddResponse)Enum.Parse(typeof(RecipientAddResponse), response.add2Return.ToString());
            }
            catch (Exception ex)
            {
                _logger.Error(ex.Message, ex);
                return null;
            }
        }

        /// <summary>
        /// Parse profile values and return keys and values as two separate arrays
        /// </summary>
        /// <param name="profileValues"></param>
        /// <returns></returns>
        private Tuple<string[], string[]> ParseProfileValues(Dictionary<string, string> profileValues)
        {
            List<string> keys = new List<string>();
            List<string> values = new List<string>();

            foreach (var value in profileValues)
            {
                keys.Add(value.Key);
                values.Add(value.Value);
            }

            return new Tuple<string[], string[]>(keys.ToArray(), values.ToArray());
        }
    }

}